'use strict';
const _ = require('lodash');
const async = require('async');

// Please see https://redoxengine.atlassian.net/wiki/spaces/CoreTransactionServices/pages/856424449/Bigints+in+PG
// for why we do this
const types = require('pg').types;
types.setTypeParser(20 /*bigint oid */ , 'text', n => parseInt(n, 10));

// Require the base adapter we plan to extend
const baseAdapter = require('./vendor/adapter');

module.exports = (function () {
  // Make our own copy of the adapter to modify
  const adapter = _.cloneDeep(baseAdapter);
  adapter.identity = 'sails-redox-postgresql';

  const _connections = {}; // Keep track of model references in private state

  // Asynchronous identity function -- used when no afterFind is available for model.
  // Defined as an internal function so that methods can detect it and short-circuit processing.
  const identity = (values, cb) => cb(null, values);

  /**
   * Given a connection and table name, looks up the Waterline collection definition's
   * afterFind() lifecycle hook. If no hook is specified, returns an identity function.
   * 
   * @param connectionName {string} Connection name
   * @param table {string} Table name
   * @returns afterFind lifecycle callback function or the identity if none is specified.
   */
  const afterFindForTable = function afterFindForTable(connectionName, table) {
    return _.get(_connections, [connectionName, table, 'afterFind'], identity);
  };

  /**
   * Registers a new connection. Tracks collection information in local state, and delegates logic to sails-postgresql.
   * 
   * @param connection {object} Connection object
   * @param collections {List<object>} List of collection objects
   * @param cb Callback function
   */
  adapter.registerConnection = function redoxPostgresql$registerConnection(connection, collections, cb) {
    baseAdapter.registerConnection(connection, collections, (err, data) => {

      // I believe that for some versions, this parameter can be an array. We don't support that.
      if (_.isArrayLike(collections)) {
        return cb(new Error(`sails-redox-postgresql.registerConnection error registering connection ${connection.identity}: 'collections' parameter is an array, expected object`));
      }

      // cache the connection's collections for reference in find() and join().
      // Performed after the baseAdapter.registerConnection() call, because that
      // will throw better errors if connection/collection are missing.
      _connections[connection.identity] = collections;

      cb(err, data);
    });
  };

  /**
   * Finds a list of records. Delegates logic to sails-postgresql. After processing is complete, the collection's
   * afterEach lifecycle hook is run on each object returned.
   * The value provided in the lifecycle hook callback replaces the existing record object.
   * 
   * @param connectionName {string} Connection name
   * @param table {string} Table name (a.k.a. model name)
   * @param options {object} find() query options
   * @param cb Callback function
   */
  adapter.find = function redoxPostgresql$find(connectionName, table, options, cb) {
    const afterFind = afterFindForTable(connectionName, table);
    const adapterOptions = options._adapter || {};

    baseAdapter.find(connectionName, table, options, (err, recordList) => {
      if (err || adapterOptions.skipAfterFind || afterFind === identity) {
        // No more processing necessary
        return cb(err, recordList);
      }

      // Unlike update() or create(), find() can return multiple records.
      // The afterFind function is called with individual records, one at a time.
      async.map(recordList, afterFind, (err, newRecordList) => {
        return cb(err, newRecordList);
      });
    });
  };

  /**
   * Performs a native join to return records from multiple tables/collections at once.
   * Delegates logic to sails-postgresql. After processing is complete, each record (including
   * joined ones) will have it's collection's afterFind lifecycle hook called on it.
   * The value provided in the lifecycle hook callback replaces the existing record object.
   * 
   * @param connectionName {string} Connection name
   * @param table {string} Table name (a.k.a. model name)
   * @param options {object} find() query options
   * @param cb Callback function
   */
  adapter.join = function redoxPostgresql$join(connectionName, table, options, cb) {
    // clone options for our use later, because baseAdapter.join will mutate it.
    const beforeOptions = _.cloneDeep(options);
    const adapterOptions = beforeOptions._adapter || {};

    baseAdapter.join(connectionName, table, options, (err, recordList) => {
      if (err || adapterOptions.skipAfterFind) {
        return cb(err, recordList);
      }

      // Build an object of junction table joins, i.e. joins with junctionTable set
      // join should have parent table of form table1_attr1__table2_attr2
      const junctionTableJoins = _.reduce(beforeOptions.joins, (memo, join) => {
        if (join.junctionTable) {
          return _.set(memo, join.alias, join);
        }
        return memo;
      }, {});

      // join() returns a list of records. Each one needs to be updated with afterFind.
      async.map(recordList, (record, next) => {
        record = record || {}; // ensure record isn't null/undefined -- although that shouldn't happen.

        // Create an async.auto query based on the joined objects -- 
        // every property that represents a joined object will have
        // afterFind called on it. This is not recursive -- it only 
        // handles "populates" one level deep.
        // _.reduce is being used to transform an array of join option objects
        // into an object of form: { propertyName: asyncUpdateFunction }
        const query = _.reduce(beforeOptions.joins, (query, join) => {
          if (join.parent !== table) {
            // Don't do anything for joins that don't have the primary collection as a parent.
            return query;
          }

          let childTableName = join.child;

          const junctionTableForAlias = junctionTableJoins[join.alias];

          // If the alias has a junction table join object that matches our child table,
          // use that join's child on the other end of the junction instead of the junction table.
          // e.g. model1 ----> model2_alias2__model1_alias1 -----> model2
          //      (table)      (junction child table)              (true child table)
          if (junctionTableForAlias && junctionTableForAlias.parent === join.child) {
            childTableName = _.get(junctionTableForAlias, ['child'], childTableName);
          }

          const afterFind = afterFindForTable(connectionName, childTableName);

          if (afterFind === identity) {
            // No need to waste time processing this record if there is no afterFind
            return query;
          }

          // add an async function to run afterFidn on the 'join.alias' property
          return _.set(query, [join.alias],
            (cb) => {
              
              const currentValue = record[join.alias]; // get the current value of the joined object
              if (!currentValue) {
                return cb(); // skip calling afterFind on null/undefined fields
              }

              // look up afterFind based on the child table name specified in the join
              if (_.isArrayLike(currentValue)) {
                // If the property is a collection of model objects, call afterFind once for each
                async.map(currentValue, afterFind, cb);
              } else {
                afterFind(currentValue, cb);
              }
            });
        }, {});
        async.auto(query, (err, queryResults) => {
          if (err) {
            return next(err);
          }
          const afterFind = afterFindForTable(connectionName, table);
          afterFind(record, (err, record) => {
            next(err, _.assign(record, queryResults));
          });
        });
      }, (err, recordList) => {
        cb(err, recordList);
      });
    });
  };

  // Expose adapter definition
  return adapter;

})();