'use strict';
const sinon = require('sinon');
const _ = require('lodash');

const chai = require('chai');
chai.use(require('sinon-chai'));
const expect = chai.expect;


const Adapter = require('../../');
const baseAdapter = require('sails-postgresql');

describe('join', function () {
  let testRecordList, options, connection, collections, testTable, testTable2, testTable3;

  before(done => {
    sinon.stub(baseAdapter, 'registerConnection').callsArgWith(2, null);
    connection = {
      identity: 'joinTestConnection'
    };
    testTable = {
      id: 'testTable'
    };
    testTable2 = {
      id: 'testTable2'
    };
    testTable3 = {
      id: 'testTable3'
    };
    collections = {testTable, testTable2, testTable3};
    Adapter.registerConnection(connection, collections, done);
  });

  after(() => {
    baseAdapter.registerConnection.restore();
  });

  beforeEach(() => {
    options = {
      select: ['id'],
      joins: [{
        parent: 'testTable',
        alias: 'subRecord',
        child: 'testTable2'
      }, {
        parent: 'testTable_subRecordList__testTable3__arecord',
        alias: 'subRecordList',
        child: 'testTable3',
        junctionTable: true
      }, {
        parent: 'testTable',
        alias: 'subRecordList',
        child: 'testTable_subRecordList__testTable3__arecord'
      }]
    };
    testRecordList = [{
      id: 1,
      subRecord: {
        id: 321
      },
      subRecordList: [{
        id: 123
      }]
    }, {
      id: 2,
      subRecord: null,
      subRecordList: [{
        id: 12
      }, {
        id: 34
      }]
    }, {
      id: 1,
      subRecord: {
        id: 321
      },
      subRecordList: []
    }];
    sinon.stub(baseAdapter, 'join').callsArgWith(3, null, testRecordList);
    testTable.afterFind = sinon.stub().callsFake((v, cb) => cb(null, v));
    testTable2.afterFind = sinon.stub().callsFake((v, cb) => cb(null, v));
    testTable3.afterFind = sinon.stub().callsFake((v, cb) => cb(null, v));
  });

  afterEach(() => {
    baseAdapter.join.restore();
  });

  it('should call baseAdapter.join and callback with the response', function (done) {
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      expect(baseAdapter.join).to.be.calledWith('joinTestConnection', 'testTable', options, sinon.match.func);
      expect(err).not.to.exist;
      expect(recordList).to.eql(testRecordList);
      done();
    });
  });

  it('should call collection afterFind lifecycle hook on each record and joined record', function (done) {
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      _.map(recordList, rec => {
        expect(testTable.afterFind).calledWith(rec, sinon.match.func).calledAfter(baseAdapter.join);

        if (rec.subRecord) {
          expect(testTable2.afterFind).calledWith(rec.subRecord, sinon.match.func).calledAfter(baseAdapter.join);
        } else {
          expect(testTable2.afterFind).not.to.be.calledWith(rec.subRecord, sinon.match.func);
        }

        _.map(rec.subRecordList, (subRec) => 
          expect(testTable3.afterFind).calledWith(subRec, sinon.match.func).calledAfter(baseAdapter.join));

      });
      expect(err).not.to.exist;
      expect(recordList).to.eql(testRecordList);
      done();
    });
  });

  it('should not call afterFind for joins without base table as parent', function (done) {
    options.joins[0].parent = 'testTable0';
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      _.map(recordList, rec => {
        expect(testTable.afterFind).calledWith(rec, sinon.match.func).calledAfter(baseAdapter.join);

        expect(testTable2.afterFind).not.to.be.called;


        _.map(rec.subRecordList, (subRec) => 
          expect(testTable3.afterFind).calledWith(subRec, sinon.match.func).calledAfter(baseAdapter.join));

      });
      expect(err).not.to.exist;
      expect(recordList).to.eql(testRecordList);
      done();
    });
  });

  it('should not call afterFind for joins when junction table has wrong parent', function (done) {
    options.joins[1].parent = 'testTable_subRecordList_asdf_testTable3__arecord'; 
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      _.map(recordList, rec => {
        expect(testTable.afterFind).calledWith(rec, sinon.match.func).calledAfter(baseAdapter.join);

        if (rec.subRecord) {
          expect(testTable2.afterFind).calledWith(rec.subRecord, sinon.match.func).calledAfter(baseAdapter.join);
        } else {
          expect(testTable2.afterFind).not.to.be.calledWith(rec.subRecord, sinon.match.func);
        }

        expect(testTable3.afterFind).not.to.be.called;

      });
      expect(err).not.to.exist;
      expect(recordList).to.eql(testRecordList);
      done();
    });
  });

  it('should not call afterFind for joins when junction table is missing', function (done) {
    delete options.joins[1].junctionTable;
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      _.map(recordList, rec => {
        expect(testTable.afterFind).calledWith(rec, sinon.match.func).calledAfter(baseAdapter.join);

        if (rec.subRecord) {
          expect(testTable2.afterFind).calledWith(rec.subRecord, sinon.match.func).calledAfter(baseAdapter.join);
        } else {
          expect(testTable2.afterFind).not.to.be.calledWith(rec.subRecord, sinon.match.func);
        }

        expect(testTable3.afterFind).not.to.be.called;

      });
      expect(err).not.to.exist;
      expect(recordList).to.eql(testRecordList);
      done();
    });
  });

  it('should support multiple fields joining on the same table', function (done) {
    options.joins.push({
      parent: 'testTable_subRecordList__testTable3__arecord',
      alias: 'subRecordList2',
      child: 'testTable3',
      junctionTable: true
    });
    options.joins.push({
      parent: 'testTable',
      alias: 'subRecordList2',
      child: 'testTable_subRecordList__testTable3__arecord'
    });
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      expect(err).not.to.exist;
      _.map(recordList, rec => {
        expect(testTable.afterFind).calledWith(rec, sinon.match.func).calledAfter(baseAdapter.join);

        if (rec.subRecord) {
          expect(testTable2.afterFind).calledWith(rec.subRecord, sinon.match.func).calledAfter(baseAdapter.join);
        } else {
          expect(testTable2.afterFind).not.to.be.calledWith(rec.subRecord, sinon.match.func);
        }

        _.map(rec.subRecordList, (subRec) => 
          expect(testTable3.afterFind).calledWith(subRec, sinon.match.func).calledAfter(baseAdapter.join));
        _.map(rec.subRecordList2, (subRec) => 
          expect(testTable3.afterFind).calledWith(subRec, sinon.match.func).calledAfter(baseAdapter.join));

      });
      done();
    });
  });

  it('should error and not call afterFind if baseAdapter.join errors', function (done) {
    baseAdapter.join.callsArgWith(3, 'testError', testRecordList);
    Adapter.join('joinTestConnection', 'testTable', options, (err, recordList) => {
      expect(baseAdapter.join).calledWith('joinTestConnection', 'testTable', options, sinon.match.func);
      expect(err).to.eql('testError');
      expect(recordList).to.eql(testRecordList);
      expect(testTable.afterFind).not.to.be.called;
      expect(testTable2.afterFind).not.to.be.called;
      expect(testTable3.afterFind).not.to.be.called;
      done();
    });
  });
});